use std::io;
use tokio::io::AsyncBufReadExt;
use tokio::io::AsyncWriteExt;
use tokio::io::BufReader;
use tokio::net::{TcpListener, TcpStream};
use tokio::sync::broadcast::{self, Receiver, Sender};
use tokio::{select, spawn};
use uuid::Uuid;

// TODO:
// * use the log crate

#[derive(Clone, Debug)]
struct Message {
    msg: String,
    sender: String,
}

#[derive(Clone, Debug)]
enum BroadcastMessage {
    Message(Message),
    // we can add more types here, for example a graceful shutdown message
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> io::Result<()> {
    let listener = TcpListener::bind("127.0.0.1:9090").await?;
    let (br_send, _) = broadcast::channel::<BroadcastMessage>(1024);

    loop {
        let (socket, _) = listener.accept().await?;
        let (tx, rx) = (br_send.clone(), br_send.subscribe());
        spawn(async move {
            if let Err(e) = connection_handler(socket, tx, rx).await {
                println!("{:?}", e);
            }
        });
    }
}

async fn connection_handler(
    mut sock: TcpStream,
    br_sender: Sender<BroadcastMessage>,
    mut br_receiver: Receiver<BroadcastMessage>,
) -> anyhow::Result<()> {
    println!("new client joined");
    let id = Uuid::new_v4().to_hyphenated().to_string();
    println!("ID: {}", id);
    let mut nick = String::from(id.split("-").collect::<Vec<&str>>()[0]);
    // todo: change this to a broadcast message, see line 22
    br_sender.send(BroadcastMessage::Message(Message {
        msg: format!("{} has joined the chat\n", nick),
        sender: id.clone(),
    }))?;
    let (rx, mut tx) = sock.split();
    let mut bufreader = BufReader::new(rx);
    loop {
        let mut buf = String::new();

        select! {
            o = bufreader.read_line(&mut buf) => {
                let bytes = o?;
                if bytes == 0 {
                    break;
                }
                if buf.starts_with("MSG") {
                    // todo: change this to a broadcast message, see line 22
                    br_sender.send(BroadcastMessage::Message(Message { msg: format!("{} {}", nick, buf.clone()) , sender: id.clone() }))?;
                } else if buf.starts_with("NICK") {
                    let old_nick = nick.clone();
                    nick = buf[5..].trim().to_string();
                    // todo: change this to a broadcast message, see line 22
                    br_sender.send(BroadcastMessage::Message(Message { msg: format!("{} {}", old_nick, buf.clone()), sender: id.clone()}))?;
                }
            },
            o = br_receiver.recv() => {
                let msg = o?;
                match msg {
                    BroadcastMessage::Message(m) => {
                        if m.sender != id {
                            tx.write_all(m.msg.as_bytes()).await?;
                        }
                    }
                }
            }
        }
    }
    // todo: change this to a broadcast message, see line 22
    br_sender.send(BroadcastMessage::Message(Message {
        msg: format!("{} has disconnected\n", nick),
        sender: id.clone(),
    }))?;
    println!("Client {} disconnected", id);
    Ok(())
}
